--[[
Copyright (c) 2012-2013 Kaarle Ritvanen
See LICENSE file for license details
--]]

local M = {}

local TreeNode = require('acf2.model.field').TreeNode
local node = require('acf2.model.node')
local object = require('acf2.object')
local pth = require('acf2.path')
local setdefaults = require('acf2.util').setdefaults


M.Set = object.class(require('acf2.model.node').List)

function M.Set:init(context, params)
   assert(not object.isinstance(params.field, TreeNode))
   params.field.dereference = false

   object.super(self, M.Set):init(context, params, 'set')

   local function find(value)
      value = node.BoundMember(
	 self, pth.wildcard, params.field
      ):normalize(value)

      for i, member in node.pairs(self) do
	 if member == value then return i, value end
      end
   end

   local mt = getmetatable(self)

   function mt.get(k, options)
      options = setdefaults(options or {}, {dereference=true})
      local i, v = find(k)
      if i then return mt.load(i, options) end
      if options.create then return v end
   end

   function mt.__newindex(t, k, v)
      assert(v == nil)
      local i = find(k)
      if not i then return end
      mt.save(i, nil)
   end

   function mt.contains(v) return find(v) and true or false end

   local insert = mt.insert
   function mt.insert(v) if not mt.contains(v) then insert(v) end end
end


return M
